(* The generic invoker *)
let rec invoke v flow = v.(flow.(0)) (invoke v) flow

generic (+) =
| int -> int -> int => (+)
| float -> float -> float => (+.)
| string -> string -> string => (^)
| $a list -> $a list -> $a list => (@)
;;

generic (-) =
| int -> int -> int => (-)
| float -> float -> float => (-.)
;;

generic ( * ) =
| int -> int -> int => ( * )
| float -> float -> float => ( *. )
;;

generic (/) =
| int -> int -> int => (/)
| float -> float -> float => (/.)
;;

generic length = 
| string     -> int => String.length
| $a list    -> int => List.length
| $a array   -> int => Array.length
| $a Queue.t -> int => Queue.length
| $a Stack.t -> int => Stack.length
;;

generic rec print =
| string -> unit => print_string
| int -> unit => print_int
| float -> unit => print_float
| char -> unit => print_char
| $a list -> unit => 
    function 
      |	[] -> ()
      |	x :: xs -> print x; print xs 
;;

generic rec vprint =
| unit => ()
| $a -> $b => fun x -> print x; vprint
;;

generic rec flat =
| $a list list -> $b list =>
    fun l -> flat (List.flatten l)
| $a list -> $a list =>
    fun x -> x
;;

generic rec curry =
| ($a * $b -> $c) -> $d -> $e =>
    (* the input type is tuple
       but the return type is a function and not still curried *)
    fun f -> curry (fun x y -> f (x,y))
| ($a -> $b -> $c) -> $a -> $d -> $e => 
    (* the input type is done, 
       but the return type is a function and not still curried *)
    fun f x -> curry (f x)
| ($a -> $b) -> $a -> $b => 
    (* the input and return types are already curried *)
    fun f -> f
;;
(* Generic.curry (fun x u (y,(z,w)) -> x +u+ y + z +w);; *)
(* does not return int -> int -> int -> int -> int -> int
   because of the restriction of polymorphic recursion *)
